/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package Interest;

import java.awt.*;
import java.util.*;
import java.io.*;
import hamlet.*;


class Plot extends Canvas {

    private final double defaultLowY = 100;
    private final double defaultHighY = 162.88946;

    private double highX = 10;
    private double lowX = 0;
    private double highY = 20;
    private double lowY = 0;

    private int height;
    private int width;

    private int offset;
    private int offsetX;
    private int offsetY;

    private Hashtable table = new Hashtable();

    private Image backImage;
    private Color frontColor = null;
    
    private NumLabel xLabel = null;
    private NumLabel yLabel = null;

    public void reset (int years) {
	backImage = null;
	frontColor = null;
	table.clear();
	highX = years;
	repaint();
    }
    


    //****************************************************
    // Add (or remove if "investment" is null) a new investment
    // to be graphed.  The "color" is used to graph it, and it
    // replaces any existing investment of that color.
    //****************************************************

    public void addGraph (Invest investment, Color color) {
	backImage = null;
	if (investment == null) {
	    table.remove(color);
	    frontColor = null;
	    if (!table.isEmpty())
		frontColor = (Color) table.keys().nextElement();
	}
	else {
	    table.put(color, investment);
	    frontColor = color;
	}
	repaint();
    }



    public void update (Graphics g) {
	paint(g);
    }
    

    public void setYears (int years) {
	highX = years;
	backImage = null;
	repaint();
    }

    public Dimension preferredSize () {
	return(new Dimension(350,350));
    }
    

    //****************************************************
    // Paint to graphics context "g".  The image "back"
    // contains the axes and all but the most recently
    // used graph.  If the most recently used graph is
    // the one that has changed, and if the axes do not need
    // to change, then the most recently used graph can be
    // drawn on top of the image.  Otherwise, everything must
    // be done from scratch.
    //****************************************************

    public void paint (Graphics g) {

	// Determine the high and low bounds of the y-axis.  If
	// the return value is true, the bounds have changed.

	if (findBounds()) backImage = null;


	// If "backImage" is null we must recreate the background image.

	if (backImage == null) {

	    // First create an empty image

	    backImage = createImage(size().width, size().height);
	    Graphics backGraphics = backImage.getGraphics();
	    backGraphics.setFont(getFont());
	    backGraphics.setColor(getBackground());
	    backGraphics.fillRect(0, 0, size().width, size().height);

	    // Then draw the axes

	    drawAxes(backGraphics);

	    // Then draw everything except for the front color

	    Enumeration keys = table.keys();
	    while (keys.hasMoreElements()) {
		Color color = (Color) keys.nextElement();
		if (color != frontColor) drawInvestment(color, backGraphics);
	    }
	}


	// We have a valid background image.  We will display it via
	// double-buffering.  We copy the background image and then
	// draw the front-color graph over it.

	Image offImage = createImage(size().width, size().height);
	Graphics offGraphics = offImage.getGraphics();
	offGraphics.drawImage(backImage, 0, 0, this);
	drawInvestment(frontColor, offGraphics);
	

	// We display the offscreen image over the displayed
	// graphics context.

	g.drawImage(offImage, 0, 0, this);


	// Finally, we display the coordinates if they exist.

	if (xLabel != null) xLabel.paint(g);
	if (yLabel != null) yLabel.paint(g);

    }
    



    boolean findBounds () {

	// Remember the old bounds

	double oldHighY = highY;
	double oldLowY = lowY;


	// Calculate the new bounds

	highY = Double.MIN_VALUE;
	lowY = Double.MAX_VALUE;
	
	Enumeration elements = table.elements();
	while (elements.hasMoreElements()) {
	    Invest investment = (Invest) elements.nextElement();
	    highY = Math.max(highY, investment.compute(highX));
	    lowY = Math.min(lowY, investment.compute(lowX));
	}
	if (highY == Double.MIN_VALUE) highY = defaultHighY;
	if (lowY == Double.MAX_VALUE) lowY = defaultLowY;


	// Return whether or not the bounds changed.

	if (oldHighY != highY || oldLowY != lowY) {
	    xLabel = yLabel = null;
	    return(true);
	}

	return(false);

    }


    void drawAxes (Graphics g) {

	// Round off the bounds and create string equivalents

	double roundedLowY = Math.ceil(lowY);
	double roundedHighY = Math.floor(highY);

	String labelLowY = Double.toString(roundedLowY);
	String labelHighY = Double.toString(roundedHighY);
	String labelLowX = Double.toString(lowX);
	String labelHighX = Double.toString(highX);


	// Determine offsets and widths

	FontMetrics fm = getFontMetrics(getFont());

	offsetX = 10 + Math.max(fm.stringWidth(labelLowY),
				    fm.stringWidth(labelHighY));
	
	offsetY = 10 + fm.getHeight() + fm.getLeading();

	offset = 2 + Math.max(fm.getAscent()/2, fm.stringWidth(labelHighX)/2);

	width = size().width - offsetX - offset;
	height = size().height - offsetY - offset;


	// Draw the axes and labels

	g.setColor(Color.black);
	g.drawLine(transX(lowX), transY(lowY), transX(lowX), transY(highY));
	g.drawLine(transX(lowX), transY(lowY), transX(highX), transY(lowY));

	g.drawString(labelLowY, offsetX - 9 - fm.stringWidth(labelLowY),
		     transY(roundedLowY) + fm.getAscent()/2);

	g.drawString(labelHighY, offsetX - 9 - fm.stringWidth(labelHighY),
		     transY(roundedHighY) + fm.getAscent()/2);

	g.drawLine(offsetX-5, transY(roundedLowY),
		   offsetX, transY(roundedLowY));

	g.drawLine(offsetX-5, transY(roundedHighY),
		     offsetX+5, transY(roundedHighY));

	g.drawString(labelLowX,
		     offsetX - fm.stringWidth(labelLowX)/2,
		     size().height - fm.getLeading() - fm.getDescent());
	
	g.drawString(labelHighX,
		     transX(highX) - fm.stringWidth(labelHighX)/2,
		     size().height - fm.getLeading() - fm.getDescent());

	g.drawLine(transX(0), size().height-offsetY,
		   transX(0), size().height-offsetY+5);

	g.drawLine(transX(highX), size().height-offsetY-5,
		   transX(highX), size().height-offsetY+5);

    }


    void drawInvestment (Color color, Graphics g) {

	// If "color" is null, there's nothing to do.

	if (color == null) return;


	// Obtain the keyed investment from the hash table.

	Invest investment = (Invest) table.get(color);


	// Draw the investment by setting the color ...

	g.setColor(color);

	// Figuring out how many steps to compute ...
	
	int steps = investment.steps(highX);

	// Keep track of points ...

	double x = lowX;
	double deltaX;
	int newX, oldX, newY, oldY;

	// Continuous interest.

	if (steps == 0) {
	    oldX = transX(x);
	    oldY = transY(investment.compute(x));
	    steps = width/2;
	    deltaX = highX/steps;
	    while (steps-- > 0) {
		newX = transX(x+deltaX);
		newY = transY(investment.compute(x+deltaX));
		g.drawLine(oldX,
			   oldY,
			   newX,
			   newY);
		x += deltaX;
		oldX = newX;
		oldY = newY;
	    }
	}

	// Simple or compound interest.

	else {
	    oldX = transX(x);
	    oldY = transY(investment.compute(x));
	    deltaX = highX/steps;
	    while (steps-- > 0) {
		newX = transX(x+deltaX);
		newY = transY(investment.compute(x+deltaX));
		g.drawLine(oldX, oldY, newX, oldY);
		g.drawLine(newX, oldY, newX, newY);
		x += deltaX;
		oldX = newX;
		oldY = newY;
	    }
	}
    }
    

    private int transX (double x) {
	return(offsetX + (int) Math.round(width * x/highX));
    }

    private int transY (double y) {
	return(offset + (int) Math.round(height * (highY-y)/(highY-lowY)));
    }


    public void resize (int w, int h) {
	backImage = null;
	xLabel = yLabel = null;
	super.resize(w, h);
    }


    public void resize (Dimension d) {
	backImage = null;
	xLabel = yLabel = null;
	super.resize(d);
    }

    public void reshape (int x, int y, int w, int h) {
	backImage = null;
	xLabel = yLabel = null;
	super.reshape(x,y,w,h);
    }


    public boolean mouseDown (Event evt, int x, int y) {

	if (x < offsetX || x > offsetX + width ||
	    y < offset || y > offset + height) return(true);
         
	FontMetrics fm = getFontMetrics(getFont());
	
	if (xLabel == null) 
	    xLabel = new NumLabel(this, 0, "y", size().width-2,
				  transY(lowY) - 2*fm.getHeight() - 5,
				  Label.RIGHT);
	else
	    xLabel.move(size().width-2,
			transY(lowY) - 2*fm.getHeight() - 5);

	if (yLabel == null)
	    yLabel = new NumLabel(this, 0, "$", size().width-2,
				  transY(lowY) - fm.getHeight() - 5,
				  Label.RIGHT);
	else
	    yLabel.move(size().width-2,
			transY(lowY) - fm.getHeight() - 5);

	return(mouseDrag(evt, x, y));

    }
    
    public boolean mouseDrag (Event evt, int x, int y) {

	if (x < offsetX || x > offsetX + width ||
	    y < offset || y > offset + height) return(true);
         
	double xValue = highX -
	                (highX-lowX) *
			(size().width - offset - x) / (double) width;

	double yValue = lowY +
			(highY-lowY) *
			(size().height - offsetY - y) / (double) height;

	xLabel.setNumber(xValue);
	yLabel.setNumber(yValue);

	return(true);
    }
    


}


